View Source Pythonx.Object (Pythonx v0.4.9)
A struct holding a Python object.
This is an opaque struct used to pass Python objects around in Elixir code.
A struct holding a Python object.
This is an opaque struct used to pass Python objects around in Elixir code.